/**********************************************************************
*
*	DXL:		lib.inc
*	Instance:		1
*	Description:	Synergy/DOORS Integration
*			General purpose functions used by the DOORS CM Synergy Integration
*	%created_by:	usmave %
*	%date_created:	Wednesday, October 28, 2009 8:05:05 AM %
*
*   Copyright (c) 2001 Telelogic AB
**********************************************************************/

// sets the dxl timeout interval to a longer cycles
pragma runLim, 700000000

#include <addins/synergy/Utils/xml_utils.inc>

void debugRecTimeTable(string msg, Skip recTimeTable)
{
	if ECPSDEBUGON then
	{
		csintDebug_(msg)
		Date recTime = null
		for recTime in recTimeTable do
		{
			string sId = (string key recTimeTable)
			csintDebug_(sId "\t" recTime "")
		}
	}
}

//=================================================================
Skip cleanTwoLevelSkip(Skip aSkip)
{
	if !null aSkip then
	{
		Skip entrySkip
		for entrySkip in aSkip do
		{
			delete entrySkip
		}
		delete aSkip
	}

	return null
}

//=================================================================
Skip cleanThreeLevelSkip(Skip aSkip)
{
	if !null aSkip then
	{
		Skip entrySkip
		for entrySkip in aSkip do
		{
			cleanTwoLevelSkip(entrySkip)
		}
		delete aSkip
	}

	return null
}

//====================================================
// This is a expensive object look up
// for object cannot be accessed through object(intOf(objAbsNo)) perm
// for example, the table header and row objects.
//====================================================
Object expensiveObjectLookup(Module m, string objAbsNo)
{
	csintDebug_("expensiveObjectLookup(" objAbsNo ")")
	Object resultObject = null
	if !null m then
	{
		Object tempObj = null
		for tempObj in entire(m) do
		{
			string tempObjAbsNo = tempObj."Absolute Number" ""
			if tempObjAbsNo == objAbsNo then
			{
				resultObject = tempObj
				break
			}
		}
	}
	return resultObject
}

Object findPreObj(Object parentObj, currentObj)
{
	Object preObj = null
	if null currentObj then
	{
		csintDebug_("findPreObj(): ERROR: the pass-in current object is null")
		return null
	}

	Module m = module(currentObj)

	// First, try the DXL built-in function to see if we can get the previous siblling object
	// well, this is not working... comment it out
	// preObj = previous(sibling(currentObj))
	// if !null preObj then return preObj

	// Now, we need to try to loop through the children of the parent to see if it is really null

	Object tempObj = null
	Object o

	if null parentObj then
	{
		// null parent means that it is directly under module
		csintDebug_("findPreObj(): looping through level one objects...")
		for o in top(m) do
		{
			string oId = o."Absolute Number" ""
			string cId = currentObj."Absolute Number" ""
			if oId == cId then
			{
				preObj = tempObj
				break
			}
			tempObj = o
		}
	}
	else
	{
		csintDebug_("findPreObj(): looping through children objects of " parentObj."Absolute Number" "...")
		for o in parentObj do
		{
			string oId = o."Absolute Number" ""
			string cId = currentObj."Absolute Number" ""
			if oId == cId then
			{
				preObj = tempObj
				break
			}
			tempObj = o
		}
	}
	return preObj
}

Object findPreObj(Object currentObj)
{
	if null currentObj then
	{
		csintDebug_("findPreObj(): ERROR: the pass-in current object is null")
		return null
	}

	Object parentObj = parent(currentObj)
	return findPreObj(parentObj, currentObj)
}

Object findFirstChild(Object o)
{
	if null o then return null

	csintDebug_("findFirstChild(" o."Absolute Number" "): by looping...")

	Object childFound = null
	Object childObj
	for childObj in o do
	{
		csintDebug_("find the first child... " childObj."Absolute Number" "")
		childFound = childObj
		break
	}
	return childFound
}

Object findThisObject(Module m , string objAbsNo)
{
	if null m then return null

	Object obj = object(intOf(objAbsNo), m)
	if null obj then
	{
		obj = expensiveObjectLookup(m, objAbsNo)
	}
	return obj
}

//====================================================
// Find the section leader
//====================================================
string findSectionLeaderAbsNo(Object o, bool basedOnOpenMode)
{
	if null o then return "-1"

	Module m = module(o)
	if basedOnOpenMode then
	{
		if isEdit(m) then return "0"	// always return 0 when in exclusive edit mode
	}

	Object temp = o
	string leaderAbsNo = temp."Absolute Number" ""
	bool keepLooking = true
	while (keepLooking)
	{
		string errMsg = isAccessInherited(temp, keepLooking)
		if null errMsg then
		{
			if keepLooking then
			{
				temp = parent(temp)
				if null temp then
				{
					leaderAbsNo = "0"	// this means the leader is the module, this means the object is not in a lockable section.
					break
				}
				else
				{
					leaderAbsNo = temp."Absolute Number" ""
				}
			}
		}
		else
		{
			csintDebug_("Error: Can not check object(" leaderAbsNo ") access inherited information: " errMsg)
			break
		}
	}

	return leaderAbsNo
}

string findSectionLeaderAbsNo(Object o)
{
	return findSectionLeaderAbsNo(o, true)
}

string findSectionLeaderAbsNo(Module m, string oAbsNo)
{
	Object o = findThisObject(m, oAbsNo)
	return findSectionLeaderAbsNo(o)
}

Date getSectionLoadingTime(Module m, string sectionLeaderAbsNo, Skip recTimeTable)
{
	Date loadingTime = null
	if isEdit(m) then
	{
		if !find(recTimeTable, "0", loadingTime) then
		{
			loadingTime = null
		}
	}
	else if isShare(m) then
	{
		if !find(recTimeTable, sectionLeaderAbsNo, loadingTime) then
		{
			loadingTime = null
		}
	}
	if null loadingTime then csintDebug_("Error: can not find loading time for section(" sectionLeaderAbsNo ")")
	return loadingTime
}

Date getObjectLoadingTime(Module m, string objId, Skip recTimeTable)
{
	Date loadingTime = null
	if isEdit(m) then
	{
		loadingTime = getSectionLoadingTime(m, "0", recTimeTable)
	}
	else if isShare(m) then
	{
		string sectionLeaderAbsNo = findSectionLeaderAbsNo(m, objId)
		loadingTime = getSectionLoadingTime(m, sectionLeaderAbsNo, recTimeTable)
	}
	if null loadingTime then csintDebug_("Error: can not find loading time for object(" objId ")")
	return loadingTime
}

//====================================================
//
//====================================================
int getLockedSections(Module m, Skip &lockedSections)
{
	int numSections = 0
	Object o = null
	for o in entire(m) do
	{
		if isLockedByUser(o) then
		{
			bool thisObjHasInteritedAccess = false
			string errMsg = isAccessInherited(o, thisObjHasInteritedAccess)
			if null errMsg then
			{
				if thisObjHasInteritedAccess then
				{
				}
				else
				{
					// we found the section leader, we are going to save it into the table

					string oAbsNum = o."Absolute Number" ""
					if !find(lockedSections, oAbsNum) then
					{
						put(lockedSections, oAbsNum, oAbsNum)
						numSections++
					}
				}
			}
			else
			{
			}
		}
	}
	return numSections
}

//====================================================
// this function is used to get the size of Skip created by createString()
// it will return -1 if the skip is null
//====================================================
int getStringSkipSize(Skip aSkip)
{
	int sSize = 0
	if null aSkip then
	{
		sSize = -1
	}
	else
	{
		string sElement = null
		for	sElement in aSkip do
		{
			sSize += 1
		}
	}

	return sSize
}

//==============================================================================
//
//==============================================================================

//******************************************************************************************
// Abort the current DXL script after displaying the message

void abortMsg(string s)
{
	string msg2 = getCsintString("csString_Csint_ErrMsg_AbortPrefix")
	string msg = msg2 s
	ack msg
	halt
}


string getUniqueTempFileName(string extension)
{
	string dummyfile = null
	string fname = null
	int c1 = 0
	for (c1=0; c1<50; c1++)
	{
		dummyfile = tempFileName()
		Stream dummyout = write dummyfile
		dummyout << "dummy"
		close dummyout
 
		fname = dummyfile extension
		if (!fileExists(fname)) then break
	}
	return fname
}

string encodeXML(string source_text, bool default_report)
{
	bool follow_return_char = false
	int text_length = length(source_text)
	int i
	char ch
	Buffer b = create
	for (i=0; i<text_length; i++)
	{
		ch = source_text[i]

		// skip extra new line if \r and \n are together
		//
		if (ch=='\n')
		{
			if (follow_return_char) then 
			{
				continue
			}
		}

		// remember \r detected
		//
		if (ch=='\r')
		{
			follow_return_char = true
			// always use new line
			ch = '\n'
		}
		else
		{
			follow_return_char = false
		}

		if (ch=='\034' || ch=='\035' || ch=='\036')
		{
			continue
		}

		// encode pre-defined XML entity
		//
		if (ch=='&')
		{
			b += "&amp;"
		}
		else if (ch=='<')
		{
			b += "&lt;"
		}
		else if (ch=='>')
		{
			b += "&gt;"
		}
		else if (ch=='"')
		{
			b += "&quot;"
		}
		else if (ch=='\'')
		{
			b += "&apos;"
		}

		// normal case, just append it to buffer
		//
		else
		{
			b += ch
		}
	}
	string out_text = stringOf b
	delete b
	return out_text
}

string encodeXML(string s)
{
	return encodeXML(s, false)
}

//******************************************************************************************
// Strip all leading and trailing spaces from the string "s"

string stripSpaces(string s)
	{
  	int p1 = 0, p2 = 0

	// Strip leading spaces
	for (p1 = 0 ; p1 < length s ; p1++)
		{
		if (s[p1] != ' ' ) then
			break
		}

	// Strip trailing spaces
	for (p2 = (length s) - 1 ; p2 > p1  ; p2--)
		{
		if (s[p2] != ' ' ) then
			break
		}

	return s[p1:p2]
	}


//******************************************************************************************
// Check for the existance of the specified Project name. A case insensitive compare is
// done on the project name.

bool existsProject(string &proj)
	{
	Project projRef = null
	bool found = false

	for projRef in database do
		{
		if (cistrcmp(name(projRef), proj) == 0)
			{
			proj = name(projRef) // Make sure the actual project name case is used
			found = true
			break
			}
		}

	return found
	}


//******************************************************************************************
// Check that the specified module name exists within the given project. This function
// assumes that the specified project name is valid. A case insensitive compare is done
// on the Module name. If a module name match is found, the actual case sensitive name
// is returned.

bool existsModule(string proj, &mod)
	{
	Project p = project("/" proj)
	Item itemRef
	bool found = false

	for itemRef in p do
		{
		if (type(itemRef) == "Formal")
			{
			if (cistrcmp(name(itemRef), mod) == 0)
				{
				mod = name(itemRef)
				found = true
				break
				}
			}
		}

	return found
	}




string construct_csint_DOORS_attrName(string crAttr)
{
	Buffer b = create
	b += CSINT_DOORS_OBJECT_ATTR_PREFIX
	b += crAttr
	string csintAttr = stringOf b
	delete b
	return csintAttr
}

string retrieve_CS_attName(string doorsAttr)
{
	Buffer b = create
	b += doorsAttr
	string csAttr = b[6:]
	return csAttr
}


int parse_into_Skip(string data, char delimiter, Skip &aSkip)
{
	Regexp re = regexp "[~!@#$%^&*_+`-=|\\:;,./<>?a-zA-Z0-9]+"
	int num_of_elements = 0
	if (!null data) then
	{
		string a_element
		while (!null data)
		{
			a_element = getNextElement(data, delimiter)
			if (!null a_element) then
			{
				if (re a_element)
				{
					put(aSkip, num_of_elements, a_element[match 0])
					num_of_elements += 1
				}
			}
		}
	}
	return num_of_elements
}

int read_attrsList_into_Skip(string data, Skip &attrSkip)
{
	int num_of_attrs = 0
	if (!null data) then
	{
		string crAttr
		while (!null data)
		{
			crAttr = getNextElement(data, '|')
			if (!null crAttr) then
			{
				put(attrSkip, crAttr, crAttr)
				num_of_attrs += 1
			}
		}
	}
	return num_of_attrs
}

int read_csint_custom_attrs_into_Skip(Skip &attrSkip)
{
	int num_of_attrs = 0
	string data = (current Module).CSINT_DOORS_MODULE_CUSTOM_ATTRS
	num_of_attrs = read_attrsList_into_Skip(data, attrSkip)
	return num_of_attrs
}

bool viewAlreadyExist(Module m, string vName)
{
	bool result = false
	string existViewName
	for existViewName in views m do
	{
		if (vName == existViewName) then
		{
			result = true
			break;
		}
	}
	return result
}

bool viewAlreadyExist(string vName)
{
	return viewAlreadyExist(current Module, vName)
}



//******************************************************************************************
// Get the unique Object ID

string csintGetObjectUniqueID(string mid, string oAbsNo)
{
	string ID = "req" mid "_" oAbsNo
	return ID
}

string csintGetObjectUniqueID(Module m, string oAbsNo)
{
	Item i = item fullName m
	string mid = uniqueID i
	string ID = csintGetObjectUniqueID(mid, oAbsNo)
	return ID
}

string csintGetObjectUniqueID(Object o)
{
	string ID = ""
	if (!null o)
	{
		string oAbsNo = o."Absolute Number" 
		Module m = module o
		ID = csintGetObjectUniqueID(m, oAbsNo)
	}
	return ID
}


//******************************************************************************************
// Get the full Object name

string getObjectFullname (Object o)
	{
	string objectName = ""
	if (!null o)
		{
		string oid = o."Absolute Number"
		Module m = module o
		string prefix = m."Prefix" 
		string fullname = fullName m
		objectName = fullname "/" prefix oid
		}
	return objectName
	}

//******************************************************************************************
// Create a HTML format file for CR submission

void CR_submission (	string filename,
			string html_object_Text,
			string objectID, 
			string objectPath, 
			string CS_Host, 
			string CS_Port,
			string CS_Protocol,
			string CS_Token,
			string currentBaseline,
			string transition,
			string associatedRCR)
	{
	csintDebug_("objectID = " objectID)
	csintDebug_("objectPath = " objectPath)
	csintDebug_("html_object_Text = " html_object_Text)
	csintDebug_("CR submission write to temp file: " filename)

	string escObjPath = escapeXML(objectPath)
	int systemCodepage = currentANSIcodepage()
	Stream out
	string systemCodepageName
	if (systemCodepage == CP_LATIN1)
	{
		out = write(filename, systemCodepage)
		systemCodepageName = "iso8859-1"
	}
	else
	{
		out = write(filename, CP_UTF8)
		systemCodepageName = "UTF-8"
	}


	out << "<HTML>\n"
	out << "<!-- saved from url=(0014)about:internet -->\n"
	out << "\t<HEAD>\n"
	out << "\t\t<META HTTP-EQUIV=Content-Type content=" "\"text/html;charset=" systemCodepageName "\">\n"
	out << "\t\t<TITLE>" CSNAME "</TITLE>\n"
	out << "\n"
	out << "\t\t<SCRIPT LANGUAGE=\"JavaScript1.2\">\n"
	out << "\n"
	string CR_data
	if (null associatedRCR)
	{ //
		CR_data = "\"<EXTERNAL_CONTEXT_DATA><ATTRIBUTE NAME='problem_synopsis' MODE='IS_REQUIRED'>" escObjPath "</ATTRIBUTE><ATTRIBUTE NAME='problem_description' MODE='IS_REQUIRED'>" html_object_Text "</ATTRIBUTE><ATTRIBUTE NAME='request_type' MODE='IS_REQUIRED'>Enhancement</ATTRIBUTE><ATTRIBUTE NAME='_CREATE_MISC_OBJECT'>"objectID"</ATTRIBUTE><ATTRIBUTE NAME='_MISC_OBJECT_RELATION'>associated_req</ATTRIBUTE><ATTRIBUTE NAME='doors_submit_baseline'>" currentBaseline "</ATTRIBUTE></EXTERNAL_CONTEXT_DATA>\";"
	}
	else
	{
		CR_data = "\"<EXTERNAL_CONTEXT_DATA><ATTRIBUTE NAME='problem_synopsis' MODE='IS_REQUIRED'>" escObjPath "</ATTRIBUTE><ATTRIBUTE NAME='problem_description' MODE='IS_REQUIRED'>" html_object_Text "</ATTRIBUTE><ATTRIBUTE NAME='request_type' MODE='IS_REQUIRED'>Enhancement</ATTRIBUTE><ATTRIBUTE NAME='_CREATE_MISC_OBJECT'>"objectID"</ATTRIBUTE><ATTRIBUTE NAME='_MISC_OBJECT_RELATION'>associated_req</ATTRIBUTE><ATTRIBUTE NAME='doors_submit_baseline'>" currentBaseline "</ATTRIBUTE><ATTRIBUTE NAME='_ASSOCIATED_RCR'>" associatedRCR "</ATTRIBUTE></EXTERNAL_CONTEXT_DATA>\";"
	}
	csintDebug_("CR_data=" CR_data)
	out << "\t\t\tvar externalContextData ="CR_data"\n"
	out << "\n"
	out << "\t\t\tfunction submitRequest(externalContextData)\n"
	out << "\t\t\t{\n"
	out << "\t\t\twindow.document.request_data.EXTERNAL_CONTEXT_DATA.value = externalContextData;\n"
	out << "\t\t\twindow.document.request_data.submit();\n"
	out << "\t\t\t}\n"
	out << "\t\t</SCRIPT>\n"
	out << "\t</HEAD>\n"
	out << "\n"
	out << "\t<BODY ONLOAD=\"submitRequest(externalContextData)\">\n"
	out << "\n"
	out << "\t\t<FORM NAME=\"request_data\" METHOD=POST ACTION=\"" CS_Protocol "://"CS_Host":"CS_Port"/servlet/com.continuus.websynergy.servlet.PTweb\">\n"
	out << "\t\t\t<INPUT NAME=\"ACTION_FLAG\"           TYPE=HIDDEN VALUE=\"frameset_form\">\n"

	if (null transition) then
	{
		out << "\t\t\t<INPUT NAME=\"TEMPLATE_FLAG\"         TYPE=HIDDEN VALUE=\"ChangeSynergySubmit\">\n"
	}
	else
	{
		out << "\t\t\t<INPUT NAME=\"TEMPLATE_FLAG\"         TYPE=HIDDEN VALUE=\"ChangeSynergySubmitSpecific\">\n"
		out << "\t\t\t<INPUT NAME=\"FRAME_TEMPLATE_FLAG\"	TYPE=HIDDEN VALUE=\"" transition "\">\n"
	}
	out << "\t\t\t<INPUT NAME=\"EXTERNAL_CONTEXT_DATA\"   TYPE=HIDDEN VALUE=\"NO_ACTION\">\n"
	out << "\t\t\t<INPUT NAME=\"role\"                  TYPE=HIDDEN VALUE=\"User\">\n"
	out << "\t\t\t<INPUT NAME=\"token\"                 TYPE=HIDDEN VALUE=\"" CS_Token"\">\n"
	out << "\t\t</FORM>\n"
	out << "\t</BODY>\n"
	out << "</HTML>\n"
	close out
	}

//******************************************************************************************
// Create a HTML format file for MRR submission

void MRR_submission (	string filename,
			string CS_Host, 
			string CS_Port,
			string CS_Protocol,
			string CS_Token,
			string moduleID,
			string transition)
	{
	Stream out = write(filename, CP_UTF8)
	out << "<HTML>\n"
	out << "<!-- saved from url=(0014)about:internet -->\n"
	out << "\t<HEAD>\n"
	out << "\t\t<META HTTP-EQUIV=Content-Type content=" "\"text/html;charset=UTF-8" "\">\n"
	out << "\t\t<TITLE>" CSNAME "</TITLE>\n"
	out << "\n"
	out << "\t\t<SCRIPT LANGUAGE=\"JavaScript1.2\">\n"
	out << "\n"
	string CR_data
		CR_data = "\"<EXTERNAL_CONTEXT_DATA><ATTRIBUTE NAME='moduleID' MODE='IS_REQUIRED'>"moduleID"</ATTRIBUTE></EXTERNAL_CONTEXT_DATA>\";"
	out << "\t\t\tvar externalContextData ="CR_data"\n"
	out << "\n"
	out << "\t\t\tfunction submitRequest(externalContextData)\n"
	out << "\t\t\t{\n"
	out << "\t\t\twindow.document.request_data.EXTERNAL_CONTEXT_DATA.value = externalContextData;\n"
	out << "\t\t\twindow.document.request_data.submit();\n"
	out << "\t\t\t}\n"
	out << "\t\t</SCRIPT>\n"
	out << "\t</HEAD>\n"
	out << "\n"
	out << "\t<BODY ONLOAD=\"submitRequest(externalContextData)\">\n"
	out << "\n"
	out << "\t\t<FORM NAME=\"request_data\" METHOD=POST ACTION=\"" CS_Protocol "://"CS_Host":"CS_Port"/servlet/com.continuus.websynergy.servlet.PTweb\">\n"
	out << "\t\t\t<INPUT NAME=\"ACTION_FLAG\"           TYPE=HIDDEN VALUE=\"frameset_form\">\n"

	if (null transition) then
	{
		out << "\t\t\t<INPUT NAME=\"TEMPLATE_FLAG\"         TYPE=HIDDEN VALUE=\"ChangeSynergySubmit\">\n"
	}
	else
	{
		out << "\t\t\t<INPUT NAME=\"TEMPLATE_FLAG\"         TYPE=HIDDEN VALUE=\"ChangeSynergySubmitSpecific\">\n"
		out << "\t\t\t<INPUT NAME=\"FRAME_TEMPLATE_FLAG\"	TYPE=HIDDEN VALUE=\"" transition "\">\n"
	}
	out << "\t\t\t<INPUT NAME=\"EXTERNAL_CONTEXT_DATA\"   TYPE=HIDDEN VALUE=\"NO_ACTION\">\n"
	out << "\t\t\t<INPUT NAME=\"role\"                  TYPE=HIDDEN VALUE=\"User\">\n"
	out << "\t\t\t<INPUT NAME=\"token\"                 TYPE=HIDDEN VALUE=\"" CS_Token"\">\n"
	out << "\t\t</FORM>\n"
	out << "\t</BODY>\n"
	out << "</HTML>\n"
	close out
	}

//*****************************************************************************************
// Create a HTML format file for CR submission without context data or relation to the current DOORS object. 
// Also, this submit always submit to a specific state based on a transition template name that is passed in.
void CR_submission_no_context(	string filename,
				string CS_Host,
				string CS_Port,
				string CS_Token,
				string CS_Protocol,
				string transition)
	{
	Stream out = write(filename, CP_UTF8)
	out << "<HTML>\n"
	out << "<!-- saved from url=(0014)about:internet -->\n"
	out << "\t<HEAD>\n"
	out << "\t\t<META HTTP-EQUIV=Content-Type content=" "\"text/html;charset=UTF-8" "\">\n"
	out << "\t\t<TITLE>" CSNAME "</TITLE>\n"
	out << "\n"
	out << "\t\t<SCRIPT LANGUAGE=\"JavaScript1.2\">\n"
	out << "\n"
	out << "\t\t\tfunction submitRequest()\n"
	out << "\t\t\t{\n"
	out << "\t\t\twindow.document.request_data.submit();\n"
	out << "\t\t\t}\n"
	out << "\t\t</SCRIPT>\n"
	out << "\t</HEAD>\n"
	out << "\n"
	out << "\t<BODY ONLOAD=\"submitRequest()\">\n"
	out << "\n"
	out << "\t\t<FORM NAME=\"request_data\" METHOD=POST ACTION=\"" CS_Protocol "://" CS_Host ":" CS_Port "/servlet/com.continuus.websynergy.servlet.PTweb\">\n"
	out << "\t\t\t<INPUT NAME=\"ACTION_FLAG\"           TYPE=HIDDEN VALUE=\"frameset_form\">\n"
	out << "\t\t\t<INPUT NAME=\"TEMPLATE_FLAG\"         TYPE=HIDDEN VALUE=\"ChangeSynergySubmitSpecific\">\n"
	out << "\t\t\t<INPUT NAME=\"FRAME_TEMPLATE_FLAG\"	TYPE=HIDDEN VALUE=\"" transition "\">\n"
	out << "\t\t\t<INPUT NAME=\"EXTERNAL_CONTEXT_DATA\"   TYPE=HIDDEN VALUE=\"NO_ACTION\">\n"
	out << "\t\t\t<INPUT NAME=\"role\"                  TYPE=HIDDEN VALUE=\"User\">\n"
	out << "\t\t\t<INPUT NAME=\"token\"                 TYPE=HIDDEN VALUE=\"" CS_Token"\">\n"
	out << "\t\t</FORM>\n"
	out << "\t</BODY>\n"
	out << "</HTML>\n"
	close out
	}


//******************************************************************************************
// Create a HTML format file for CR show
// TODO  make host and port as variable
void CR_show (string filename, string CR_number, string CS_Host, string CS_Port, string CS_Token, string CS_Protocol)
	{
	Stream out = write(filename, CP_UTF8)
	out << "<HTML>\n"
	out << "<!-- saved from url=(0014)about:internet -->\n"
	out << "\t<HEAD>\n"
	out << "\t\t<META HTTP-EQUIV=Content-Type content=" "\"text/html;charset=UTF-8" "\">\n"
	out << "\t\t<TITLE>" CSNAME "</TITLE>\n"
	out << "\n"
	out << "\t<BODY ONLOAD=\"window.document.request_data.submit()\">\n"
	out << "\n"
	out << "\t\t<FORM NAME=\"request_data\" METHOD=POST ACTION=\"" CS_Protocol "://"CS_Host":"CS_Port"/servlet/com.continuus.websynergy.servlet.PTweb\">\n"
	out << "\t\t\t<INPUT NAME=\"ACTION_FLAG\"           TYPE=HIDDEN VALUE=\"frameset_form\">\n"
	out << "\t\t\t<INPUT NAME=\"TEMPLATE_FLAG\"         TYPE=HIDDEN VALUE=\"ChangeSynergyShowDetails\">\n"
	out << "\t\t\t<INPUT NAME=\"role\"                  TYPE=HIDDEN VALUE=\"User\">\n"
	out << "\t\t\t<INPUT NAME=\"token\"                 TYPE=HIDDEN VALUE=\""CS_Token"\">\n"
	out << "\t\t\t<INPUT NAME=\"problem_number\"        TYPE=HIDDEN VALUE=\""CR_number"\">\n"
	out << "\t\t</FORM>\n"
	out << "\t</BODY>\n"
	out << "</HTML>\n"
	close out
	}




//******************************************************************************************
// Convert a string to xml format like

string text_to_xml(string sin)
	{
	Buffer	sout = create
	string	szNew
	char	ch
	int	iCounter
	int	iLength	= 0
	int	iStart	= 0
	int	sinlen = length sin

	for (iCounter = 0; iCounter < sinlen; iCounter++)
		{
		bool bSpecialChar = true
		ch = sin[iCounter]

		if ( ch == '\n' ) 
			{
    			szNew = "\\n"
			}
		else if ( ch == '\r' ) 
			{
    			szNew = "\\r"
			}
		else if ( ch == '"' )
			{
			szNew = "&#034;"
			}
		else if ( ch == '$' )
			{
			szNew = "&#036;"
			}
		else if ( ch == '&' ) 
			{
			szNew = "&#038;"
			} 
		else if ( ch == '\'' )
			{
			szNew = "&#039;"
			}
		else if ( ch == '\\' )
			{
			szNew = "\\\\"
			}
		else if ( ch == '<' )
			{
			szNew = "&#060;"
			}
		else if ( ch == '>' )
			{
			szNew = "&#062;"
			}
		else
			bSpecialChar = false

		if ( bSpecialChar )
			{
			sout += szNew
			} 
		else 
			{
			sout +=  sin[ iCounter]
			}
		}

	szNew = stringOf sout
	delete sout
	return szNew
	}


//******************************************************************************************
// Pick up the token

string getCSToken()
	{
	string outfilename = doorsHome"/temp/tokenout.txt"
	string logfilename = doorsHome"/temp/tokenlog.txt"
	Stat checkfile
	int filesize = 0
	string s, CS_Token

	// make sure the outfilename and logfilename do not exist.
	deleteFile outfilename
	deleteFile logfilename

	// pick up the token
	// format the command line
	string cmd = "ccm_win cs_token -o \"" outfilename "\" -l \"" logfilename "\""
	system cmd
	
	// try to get the status handle for the filename
	// if it's null then the file is not exist yet
	checkfile = create logfilename
	// loop through the file is created
	while ((null checkfile) or (filesize == 0))
		{
		checkfile = create logfilename
		if (!null checkfile) then
			filesize = size(checkfile)
		}
	delete checkfile

	// read the content of log
	Stream loginput = read(logfilename)
	loginput >> s
	close loginput
	if (s != "0") then
		abortMsg("command '"cmd"' failed")
		
	// no match found
	if (!fileExists(outfilename))
		{
		halt
		}
	else
		{
		Stream input = read outfilename
		input >> CS_Token
		close input
		}
	// clean up temp files.
	deleteFile outfilename
	deleteFile logfilename

	return CS_Token
	}

void check_module_initialization(Module m)
{
	if null m then
	{
		string csint_ErrMsg_NullModule = getCsintString("csString_Csint_ErrMsg_NullModule")
		abortMsg(csint_ErrMsg_NullModule)
	}
	else
	{
		string mid = uniqueID(item(fullName(m)))
		if !moduleIsConfigured(mid) then
		{
			string csint_ErrMsg_PleaseConfigThisModule = getCsintString("csString_Csint_ErrMsg_PleaseConfigThisModule", CSINT_TCI_NAME)
			abortMsg(csint_ErrMsg_PleaseConfigThisModule)
		}
		else
		{
			Skip configData = createString
			readCsintConfigurationFile(mid, configData)	// this read module config file only
			string templateLink = null
			string csint_ErrMsg_PleaseDefineTemplate = getCsintString("csString_Csint_ErrMsg_PleaseDefineTemplate")
			if !find(configData, CSINT_CONFIG_USE_TEMPLATE, templateLink) then
			{
				delete configData
				abortMsg(csint_ErrMsg_PleaseDefineTemplate)
			}
			else
			{
				if null templateLink then
				{
					delete configData
					abortMsg(csint_ErrMsg_PleaseDefineTemplate)
				}
				else
				{
					if !confFileExists(CSINT_TEMPLATES_DIR "/" templateLink, templateConfArea) then
					{
						delete configData
						string csint_ErrMsg_TempleteFileNotExist = getCsintString("csString_Csint_ErrMsg_TempleteFileNotExist")
						abortMsg(csint_ErrMsg_TempleteFileNotExist)
					}
					else
					{
						// we will re-read everything here again
						delete configData
						configData = createString
						getModuleConfigurationData(m, configData)

						// check critical entries...
						string templateName = "<not defined>"	// this should not happen
						find(configData, CSINT_CONFIG_TEMPLATE_NAME, templateName)

						Buffer errBuffer = create
						string csHost = null
						string csPort = null
						string csProtocol = null
						string cmDatabase = null
						find(configData, CSINT_CONFIG_CS_HOST, csHost)
						find(configData, CSINT_CONFIG_CS_PORT, csPort)
						find(configData, CSINT_CONFIG_CS_PROTOCOL, csProtocol)
						find(configData, CSINT_CONFIG_DB_PATH, cmDatabase)
						if null csHost then
						{
							errBuffer += "\n\t"
							errBuffer += CSNAME
							errBuffer += " "
							errBuffer += getCsintString("csString_String_Host_Name")
						}
						if null csPort then
						{
							errBuffer += "\n\t"
							errBuffer += CSNAME
							errBuffer += " "
							errBuffer += getCsintString("csString_String_Port_Number")
						}
						if null csProtocol then
						{
							errBuffer += "\n\t"
							errBuffer += CSNAME
							errBuffer += " "
							errBuffer += getCsintString("csString_String_Protocol")
						}
						if null cmDatabase then
						{
							errBuffer += "\n\t"
							errBuffer += CMNAME
							errBuffer += " "
							errBuffer += getCsintString("csString_String_Database")
						}
						if length(errBuffer) > 0 then
						{
							string errMsg = stringOf errBuffer
							delete errBuffer
							delete configData
							string csint_ErrMsg_PleaseCheckTemplate = getCsintString("csString_Csint_ErrMsg_PleaseCheckTemplate", templateName, CSNAME)
							abortMsg(csint_ErrMsg_PleaseCheckTemplate errMsg)
						}
						delete errBuffer
					}
				}
			}
			delete configData
		}
	}
}

string buildStateQuery(Module m, string stateList, string defaultStates)
{
	string myStateList = stateList
	Buffer b = create
	b = "("
	if (null myStateList) then myStateList = defaultStates
	string stateName = getStringElement(myStateList, CSINT_CSDELIM_CHAR, false)

	while (!null stateName)
	{
		b += "crstatus='"
		b += stateName
		b += "'"
		stateName = getStringElement(myStateList, CSINT_CSDELIM_CHAR, false)
		if (!null stateName) then
		{
			b += " or "
		}
	}

	b += ")"
	string retVal = stringOf(b)
	delete b
	return retVal
}

string buildAssignStateQuery(Module m)
{
	string assignStateList = getCSINTConfigurationSettings(CSINT_CONFIG_RCP_ASSIGNED_STATES, m)
	return buildStateQuery(m, assignStateList, CSINT_DEFAULT_RCP_ASSIGNED_STATES)
}

string buildMrrAssignStateQuery(Module m)
{
	string assignStateList = getCSINTConfigurationSettings(CSINT_CONFIG_MRR_ASSIGNED_STATE, m)
	return buildStateQuery(m, assignStateList, CSINT_DEFAULT_MRR_ASSIGNED_STATE)
}

string buildApprovedStateQuery(Module m)
{
	string approvedStateList = getCSINTConfigurationSettings(CSINT_CONFIG_RCP_READY_FOR_APPLY_STATE, m)
	return buildStateQuery(m, approvedStateList, CSINT_DEFAULT_RCP_APPROVED_STATE)
}

string buildReviewStateQuery(Module m)
{
	string reviewStateList = getCSINTConfigurationSettings(CSINT_CONFIG_RCP_REVIEW_STATES, m)
	return buildStateQuery(m, reviewStateList, CSINT_DEFAULT_RCP_REVIEW_STATES)
}

string buildCompleteStateQuery(Module m)
{
	string completeStateList = getCSINTConfigurationSettings(CSINT_CONFIG_RCR_COMPLETE_STATE, m)
	return buildStateQuery(m, completeStateList, CSINT_DEFAULT_RCR_COMPLETE_STATE)
}

string buildAppliedStateQuery(Module m)
{
	string appliedStateList = getCSINTConfigurationSettings(CSINT_CONFIG_RCP_APPLIED_STATE, m)
	return buildStateQuery(m, appliedStateList, CSINT_DEFAULT_RCP_APPLIED_STATE)
}

string buildRcrForModuleQuery(Module m)
{
	string dbID = getDoorsDB_ID()
	Buffer b = create
	Item i = item fullName m
	string mid = uniqueID i

//	b = "has_dynamic_attrs_new(doorsID match 'req"
	b = "has_dynamic_attrs_new(doorsID ='req"
	b += mid
//	b += "*' and " CSINT_RCM_RC_DOORS_DB_ID "='" dbID "'"
	b += "' and " CSINT_RCM_RC_DOORS_DB_ID "='" dbID "'"
	b += ")"

	string retVal = stringOf(b)
	delete b
	return retVal
}

string buildRcrForObjectQuery(Object targetObject)
{
	string oid = csintGetObjectUniqueID(targetObject)
	return "has_dynamic_attrs_new(doorsID='" oid "')"
}

OleAutoObj createCSAttrObj(string attrname, string type, string label)
{
	OleAutoObj csAttrObj = oleCreateAutoObject "cslib.csObjectData"
	oleMethod(csAttrObj, "Create")
	olePut(csAttrObj, "Name", attrname)
	olePut(csAttrObj, "Type", type)
	olePut(csAttrObj, "Label", label)
	return csAttrObj
}

bool addCSObjectDatatoArray(OleAutoObj &attrArray, OleAutoObj objdata, int pos)
{
	OleAutoArgs oArgs = create
	put (oArgs, pos)
	put (oArgs, objdata)
	string errMsg = oleMethod(attrArray, "AddDataObject", oArgs)
	delete oArgs
	if (!null errMsg)
	{
		print "Unable to update attribute array : " errMsg "\n" 
		return false
	}
	return true
}


OleAutoObj createCMAttrObj(string attrname, string type, string value)
{
	OleAutoObj cmAttrObj = oleCreateAutoObject "cslib.csObjectData"
	oleMethod(cmAttrObj, "Create")
	olePut(cmAttrObj, "Name", attrname)
	olePut(cmAttrObj, "Type", type)
	olePut(cmAttrObj, "Value", value)
	return cmAttrObj
}

OleAutoObj createCMIntAttrObj(string attrname, int value)
{
	OleAutoObj cmAttrObj = oleCreateAutoObject "cslib.csObjectData"
	oleMethod(cmAttrObj, "Create")
	olePut(cmAttrObj, "Name", attrname)
	olePut(cmAttrObj, "Type", "integer")
	olePut(cmAttrObj, "Value", value)
	return cmAttrObj
}

bool recursiveHardDelete(Object o)
{
	bool result = true

	string resultCanDelete = canDelete(o)
	if !null resultCanDelete then
	{
		Object nextChild = first(o)
		while (!null nextChild)
		{
			Object thisChild = nextChild
			nextChild = next(sibling(thisChild))

			result = recursiveHardDelete(thisChild)
		}
	}

	if result == true then
	{
		resultCanDelete = canDelete(o)
		if null resultCanDelete then
		{
			if isDeleted(o) then undelete(o)
			hardDelete(o)
			flushDeletions
		}
		else
		{
			result = false
		}
	}

	return result
}

//====================================================
// a customized message box
//====================================================
int csintCustomizeMsgBoxReturnValue = -1
void customizeMsgBox_close(DB parentDB)
{
	//maybe we want to do something... but for now, just ingore the click.
	//csintCustomizeMsgBoxReturnValue = -1
	//hide parentDB
}

void customizeMsgBox_ok_zero(DB parentDB)
{
	csintCustomizeMsgBoxReturnValue = 0
}

void customizeMsgBox_ok_one(DB parentDB)
{
	csintCustomizeMsgBoxReturnValue = 1
}

int csintCustomizeMsgBox(DB parentDB, Module m, int severity, int iWidth, iHeight, string titleString, string mainMessage, secondMessage, thirdMessage, string userChoices[])
{
	DB myMsgBox
	if !null parentDB then
	{
		myMsgBox = create(parentDB, titleString, styleCentred|styleFixed)
	}
	else if !null m then
	{
		myMsgBox = create(m, titleString, styleCentred|styleFixed)
	}
	else
	{
		myMsgBox = create(titleString, styleCentred|styleFixed)
	}
	
	if !null mainMessage then label(myMsgBox, mainMessage)
	if !null secondMessage then label(myMsgBox, secondMessage)
	if !null thirdMessage then label(myMsgBox, thirdMessage)

	close(myMsgBox, false, customizeMsgBox_close)
	ok(myMsgBox, userChoices[0], customizeMsgBox_ok_zero)
	ok(myMsgBox, userChoices[1], customizeMsgBox_ok_one)
	
	realize myMsgBox
	setSize(myMsgBox, iWidth, iHeight)

	// do not ask me why the followign code.... it works... and I do not want to figure out why.... ask DOORS guys...
	ready myMsgBox
	busy myMsgBox
	ready myMsgBox
	ready myMsgBox

	block myMsgBox
	destroy myMsgBox
	return csintCustomizeMsgBoxReturnValue
}

// ==============================================================================
// The followings are generic reusable functions for progress bar
// Parent is the module, without DB in context
// ==============================================================================
DB csintGenericProgressBar_Start(Module m, int totalStep, string titleString, string msgString)
{
	// ---------------------------------------------------
	// show progress bar
	// ---------------------------------------------------
	int iSteps = totalStep
	DB myBox = null
	myBox = create(m, "Progress")
	realize myBox
	hide myBox
	progressStart(myBox, titleString, "", iSteps)
	progressStep(1)
	progressMessage(msgString)

	return myBox
}

// ==============================================================================
bool csintGenericProgressBar_NextStep(int stepNumber, bool allowCancel, string msgString)
{
	bool progressIsCancelled = false
	progressStep(stepNumber)
	if progressCancelled() then
	{
		if allowCancel then
		{
			progressIsCancelled = true
		}
		else
		{
			warningBox("This process can not be cancelled!!")
			progressMessage(msgString)
		}
	}
	else
	{
		progressMessage(msgString)
	}

	return progressIsCancelled
}

// ==============================================================================
void csintGenericProgressBar_Stop(DB myBox)
{
	progressStop()
	destroy myBox
}

// the pass-in skip must be created using "create"
int getCurrentObjectSelection(Module cur_mod, Skip selectionSkip)
{
	int iObjCount = 0
	Object o = null
	Object start = null
	Object end = null
	getSelection(cur_mod, start, end)
	int loopTerminator = 2

	if (null start)
	{
		// There wasn't a selection, only the current object
		loopTerminator = 1
		o = current
	}
	else
	{
		o = start
		if null end then
		{
			end = start
		}
		else
		{
			while (!null last end)
			{
				end = last end
			}
		}
	}

	if !null o then
	{
		while (loopTerminator != 0)
		{
			if (loopTerminator == 1)
			{
				//For some reason if o is set to current outside the loop, it gets unset once the loop starts.
				//If loopTerminator was 1 at the beginning of the loop, there was only one object selected
				o = current
				loopTerminator = 0
			}
			else if (o == end)
			{
				// This is the last object selected, so stop the loop after this run
				loopTerminator = 0
			}

			iObjCount++
			put(selectionSkip, iObjCount, o)

			if (loopTerminator != 0)
			o = next o
		}
	}

	return iObjCount
}

// ==============================================================================
bool isObjNewAtThisSession(Object aObj)
{
	csintDebug_("isObjNewAtThisSession():")
	bool myAnswer = false
	if !null aObj then
	{
		History h
		for h in session(aObj) do	// only the history records of current session
		{
			HistoryType ht = h.type
			csintDebug_("\t" ht "")
			if ht == createObject then
			{
				myAnswer = true
				break
			}
			else if ht == clipCopyObject then
			{
				myAnswer = true
				break
			}
		}
	}

	return myAnswer
}